@@ -43,6 +43,90 @@ unsigned long parse_blocksize(const char *bs)
     return r;
 }
 
+/*
+ * Parse a ^C style character specification.
+ * Returns NULL in `next' if we didn't recognise it as a control character,
+ * in which case `c' should be ignored.
+ * The precise current parsing is an oddity inherited from the terminal
+ * answerback-string parsing code. All sequences start with ^; all except
+ * ^<123> are two characters. The ones that are worth keeping are probably:
+ *   ^?		    127
+ *   ^@A-Z[\]^_	    0-31
+ *   a-z	    1-26
+ *   <num>	    specified by number (decimal, 0octal, 0xHEX)
+ *   ~		    ^ escape
+ */
+char ctrlparse(char *s, char **next)
+{
+    char c = 0;
+    if (*s != '^') {
+	*next = NULL;
+    } else {
+	s++;
+	if (*s == '\0') {
+	    *next = NULL;
+	} else if (*s == '<') {
+	    s++;
+	    c = (char)strtol(s, next, 0);
+	    if ((*next == s) || (**next != '>')) {
+		c = 0;
+		*next = NULL;
+	    } else
+		(*next)++;
+	} else if (*s >= 'a' && *s <= 'z') {
+	    c = (*s - ('a' - 1));
+	    *next = s+1;
+	} else if ((*s >= '@' && *s <= '_') || *s == '?' || (*s & 0x80)) {
+	    c = ('@' ^ *s);
+	    *next = s+1;
+	} else if (*s == '~') {
+	    c = '^';
+	    *next = s+1;
+	}
+    }
+    return c;
+}
+
+prompts_t *new_prompts(void *frontend)
+{
+    prompts_t *p = snew(prompts_t);
+    p->prompts = NULL;
+    p->n_prompts = 0;
+    p->frontend = frontend;
+    p->data = NULL;
+    p->to_server = TRUE; /* to be on the safe side */
+    p->name = p->instruction = NULL;
+    p->name_reqd = p->instr_reqd = FALSE;
+    return p;
+}
+void add_prompt(prompts_t *p, char *promptstr, int echo, size_t len)
+{
+    prompt_t *pr = snew(prompt_t);
+    unsigned char *result = snewn(len, unsigned char);
+    pr->prompt = promptstr;
+    pr->echo = echo;
+    pr->result = result;
+    pr->result_len = len;
+    p->n_prompts++;
+    p->prompts = sresize(p->prompts, p->n_prompts, prompt_t *);
+    p->prompts[p->n_prompts-1] = pr;
+}
+void free_prompts(prompts_t *p)
+{
+    size_t i;
+    for (i=0; i < p->n_prompts; i++) {
+	prompt_t *pr = p->prompts[i];
+	memset(pr->result, 0, pr->result_len); /* burn the evidence */
+	sfree(pr->result);
+	sfree(pr->prompt);
+	sfree(pr);
+    }
+    sfree(p->prompts);
+    sfree(p->name);
+    sfree(p->instruction);
+    sfree(p);
+}
+
 /* ----------------------------------------------------------------------
  * String handling routines.
  */
@@ -397,6 +481,7 @@ void *safemalloc(size_t n, size_t size)
 	p = NULL;
     } else {
 	size *= n;
+	if (size == 0) size = 1;
 #ifdef MINEFIELD
 	p = minefield_c_malloc(size);
 #else
@@ -540,3 +625,23 @@ void debug_memdump(void *buf, int len, int L)
 }
 
 #endif				/* def DEBUG */
+
+/*
+ * Determine whether or not a Config structure represents a session
+ * which can sensibly be launched right now.
+ */
+int cfg_launchable(const Config *cfg)
+{
+    if (cfg->protocol == PROT_SERIAL)
+	return cfg->serline[0] != 0;
+    else
+	return cfg->host[0] != 0;
+}
+
+char const *cfg_dest(const Config *cfg)
+{
+    if (cfg->protocol == PROT_SERIAL)
+	return cfg->serline;
+    else
+	return cfg->host;
+}
