@@ -1,6 +1,10 @@
 #ifndef PUTTY_UNIX_H
 #define PUTTY_UNIX_H
 
+#ifdef HAVE_CONFIG_H
+# include "uxconfig.h" /* Space to hide it from mkfiles.pl */
+#endif
+
 #include <stdio.h>		       /* for FILENAME_MAX */
 #include "charset.h"
 
@@ -101,7 +105,7 @@ void uxsel_input_remove(int id);
 
 /* uxcfg.c */
 struct controlbox;
-void unix_setup_config_box(struct controlbox *b, int midsession);
+void unix_setup_config_box(struct controlbox *b, int midsession, int protocol);
 
 /* gtkcfg.c */
 void gtk_setup_config_box(struct controlbox *b, int midsession, void *window);
@@ -145,4 +149,9 @@ void *sk_getxdmdata(void *sock, int *lenp);
     if (max < fd + 1) max = fd + 1; \
 } while (0)
 
+/*
+ * Exports from winser.c.
+ */
+extern Backend serial_backend;
+
 #endif
