@@ -13,7 +13,10 @@
 #include <termios.h>
 #include <pwd.h>
 #include <sys/ioctl.h>
+#include <sys/time.h>
+#ifndef HAVE_NO_SYS_SELECT_H
 #include <sys/select.h>
+#endif
 
 #define PUTTY_DO_GLOBALS	       /* actually _define_ globals */
 #include "putty.h"
@@ -22,6 +25,8 @@
 
 #define MAX_STDIN_BACKLOG 4096
 
+void *logctx;
+
 void fatalbox(char *p, ...)
 {
     va_list ap;
@@ -30,6 +35,10 @@ void fatalbox(char *p, ...)
     vfprintf(stderr, p, ap);
     va_end(ap);
     fputc('\n', stderr);
+    if (logctx) {
+        log_free(logctx);
+        logctx = NULL;
+    }
     cleanup_exit(1);
 }
 void modalfatalbox(char *p, ...)
@@ -40,6 +49,10 @@ void modalfatalbox(char *p, ...)
     vfprintf(stderr, p, ap);
     va_end(ap);
     fputc('\n', stderr);
+    if (logctx) {
+        log_free(logctx);
+        logctx = NULL;
+    }
     cleanup_exit(1);
 }
 void connection_fatal(void *frontend, char *p, ...)
@@ -50,6 +63,10 @@ void connection_fatal(void *frontend, char *p, ...)
     vfprintf(stderr, p, ap);
     va_end(ap);
     fputc('\n', stderr);
+    if (logctx) {
+        log_free(logctx);
+        logctx = NULL;
+    }
     cleanup_exit(1);
 }
 void cmdline_error(char *p, ...)
@@ -63,7 +80,8 @@ void cmdline_error(char *p, ...)
     exit(1);
 }
 
-struct termios orig_termios;
+static int local_tty = 0; /* do we have a local tty? */
+static struct termios orig_termios;
 
 static Backend *back;
 static void *backhandle;
@@ -78,6 +96,8 @@ char *platform_default_s(const char *name)
 	return dupstr(getenv("TERM"));
     if (!strcmp(name, "UserName"))
 	return get_username();
+    if (!strcmp(name, "SerialLine"))
+	return dupstr("/dev/ttyS0");
     return NULL;
 }
 
@@ -122,6 +142,8 @@ void ldisc_update(void *frontend, int echo, int edit)
     /* Update stdin read mode to reflect changes in line discipline. */
     struct termios mode;
 
+    if (!local_tty) return;
+
     mode = orig_termios;
 
     if (echo)
@@ -132,19 +154,219 @@ void ldisc_update(void *frontend, int echo, int edit)
     if (edit) {
 	mode.c_iflag |= ICRNL;
 	mode.c_lflag |= ISIG | ICANON;
+	mode.c_oflag |= OPOST;
     } else {
 	mode.c_iflag &= ~ICRNL;
 	mode.c_lflag &= ~(ISIG | ICANON);
+	mode.c_oflag &= ~OPOST;
+	/* Solaris sets these to unhelpful values */
 	mode.c_cc[VMIN] = 1;
 	mode.c_cc[VTIME] = 0;
+	/* FIXME: perhaps what we do with IXON/IXOFF should be an
+	 * argument to ldisc_update(), to allow implementation of SSH-2
+	 * "xon-xoff" and Rlogin's equivalent? */
+	mode.c_iflag &= ~IXON;
+	mode.c_iflag &= ~IXOFF;
     }
+    /* 
+     * Mark parity errors and (more important) BREAK on input.  This
+     * is more complex than it need be because POSIX-2001 suggests
+     * that escaping of valid 0xff in the input stream is dependent on
+     * IGNPAR being clear even though marking of BREAK isn't.  NetBSD
+     * 2.0 goes one worse and makes it dependent on INPCK too.  We
+     * deal with this by forcing these flags into a useful state and
+     * then faking the state in which we found them in from_tty() if
+     * we get passed a parity or framing error.
+     */
+    mode.c_iflag = (mode.c_iflag | INPCK | PARMRK) & ~IGNPAR;
 
     tcsetattr(0, TCSANOW, &mode);
 }
 
+/* Helper function to extract a special character from a termios. */
+static char *get_ttychar(struct termios *t, int index)
+{
+    cc_t c = t->c_cc[index];
+#if defined(_POSIX_VDISABLE)
+    if (c == _POSIX_VDISABLE)
+	return dupprintf("");
+#endif
+    return dupprintf("^<%d>", c);
+}
+
+char *get_ttymode(void *frontend, const char *mode)
+{
+    /*
+     * Propagate appropriate terminal modes from the local terminal,
+     * if any.
+     */
+    if (!local_tty) return NULL;
+
+#define GET_CHAR(ourname, uxname) \
+    do { \
+	if (strcmp(mode, ourname) == 0) \
+	    return get_ttychar(&orig_termios, uxname); \
+    } while(0)
+#define GET_BOOL(ourname, uxname, uxmemb, transform) \
+    do { \
+	if (strcmp(mode, ourname) == 0) { \
+	    int b = (orig_termios.uxmemb & uxname) != 0; \
+	    transform; \
+	    return dupprintf("%d", b); \
+	} \
+    } while (0)
+
+    /*
+     * Modes that want to be the same on all terminal devices involved.
+     */
+    /* All the special characters supported by SSH */
+#if defined(VINTR)
+    GET_CHAR("INTR", VINTR);
+#endif
+#if defined(VQUIT)
+    GET_CHAR("QUIT", VQUIT);
+#endif
+#if defined(VERASE)
+    GET_CHAR("ERASE", VERASE);
+#endif
+#if defined(VKILL)
+    GET_CHAR("KILL", VKILL);
+#endif
+#if defined(VEOF)
+    GET_CHAR("EOF", VEOF);
+#endif
+#if defined(VEOL)
+    GET_CHAR("EOL", VEOL);
+#endif
+#if defined(VEOL2)
+    GET_CHAR("EOL2", VEOL2);
+#endif
+#if defined(VSTART)
+    GET_CHAR("START", VSTART);
+#endif
+#if defined(VSTOP)
+    GET_CHAR("STOP", VSTOP);
+#endif
+#if defined(VSUSP)
+    GET_CHAR("SUSP", VSUSP);
+#endif
+#if defined(VDSUSP)
+    GET_CHAR("DSUSP", VDSUSP);
+#endif
+#if defined(VREPRINT)
+    GET_CHAR("REPRINT", VREPRINT);
+#endif
+#if defined(VWERASE)
+    GET_CHAR("WERASE", VWERASE);
+#endif
+#if defined(VLNEXT)
+    GET_CHAR("LNEXT", VLNEXT);
+#endif
+#if defined(VFLUSH)
+    GET_CHAR("FLUSH", VFLUSH);
+#endif
+#if defined(VSWTCH)
+    GET_CHAR("SWTCH", VSWTCH);
+#endif
+#if defined(VSTATUS)
+    GET_CHAR("STATUS", VSTATUS);
+#endif
+#if defined(VDISCARD)
+    GET_CHAR("DISCARD", VDISCARD);
+#endif
+    /* Modes that "configure" other major modes. These should probably be
+     * considered as user preferences. */
+    /* Configuration of ICANON */
+#if defined(ECHOK)
+    GET_BOOL("ECHOK", ECHOK, c_lflag, );
+#endif
+#if defined(ECHOKE)
+    GET_BOOL("ECHOKE", ECHOKE, c_lflag, );
+#endif
+#if defined(ECHOE)
+    GET_BOOL("ECHOE", ECHOE, c_lflag, );
+#endif
+#if defined(ECHONL)
+    GET_BOOL("ECHONL", ECHONL, c_lflag, );
+#endif
+#if defined(XCASE)
+    GET_BOOL("XCASE", XCASE, c_lflag, );
+#endif
+    /* Configuration of ECHO */
+#if defined(ECHOCTL)
+    GET_BOOL("ECHOCTL", ECHOCTL, c_lflag, );
+#endif
+    /* Configuration of IXON/IXOFF */
+#if defined(IXANY)
+    GET_BOOL("IXANY", IXANY, c_iflag, );
+#endif
+    /* Configuration of OPOST */
+#if defined(OLCUC)
+    GET_BOOL("OLCUC", OLCUC, c_oflag, );
+#endif
+#if defined(ONLCR)
+    GET_BOOL("ONLCR", ONLCR, c_oflag, );
+#endif
+#if defined(OCRNL)
+    GET_BOOL("OCRNL", OCRNL, c_oflag, );
+#endif
+#if defined(ONOCR)
+    GET_BOOL("ONOCR", ONOCR, c_oflag, );
+#endif
+#if defined(ONLRET)
+    GET_BOOL("ONLRET", ONLRET, c_oflag, );
+#endif
+
+    /*
+     * Modes that want to be set in only one place, and that we have
+     * squashed locally.
+     */
+#if defined(ISIG)
+    GET_BOOL("ISIG", ISIG, c_lflag, );
+#endif
+#if defined(ICANON)
+    GET_BOOL("ICANON", ICANON, c_lflag, );
+#endif
+#if defined(ECHO)
+    GET_BOOL("ECHO", ECHO, c_lflag, );
+#endif
+#if defined(IXON)
+    GET_BOOL("IXON", IXON, c_iflag, );
+#endif
+#if defined(IXOFF)
+    GET_BOOL("IXOFF", IXOFF, c_iflag, );
+#endif
+#if defined(OPOST)
+    GET_BOOL("OPOST", OPOST, c_oflag, );
+#endif
+
+    /*
+     * We do not propagate the following modes:
+     *  - Parity/serial settings, which are a local affair and don't
+     *    make sense propagated over SSH's 8-bit byte-stream.
+     *      IGNPAR PARMRK INPCK CS7 CS8 PARENB PARODD
+     *  - Things that want to be enabled in one place that we don't
+     *    squash locally.
+     *      IUCLC
+     *  - Status bits.
+     *      PENDIN
+     *  - Things I don't know what to do with. (FIXME)
+     *      ISTRIP IMAXBEL NOFLSH TOSTOP IEXTEN
+     *      INLCR IGNCR ICRNL
+     */
+
+#undef GET_CHAR
+#undef GET_BOOL
+
+    /* Fall through to here for unrecognised names, or ones that are
+     * unsupported on this platform */
+    return NULL;
+}
+
 void cleanup_termios(void)
 {
-    tcsetattr(0, TCSANOW, &orig_termios);
+    if (local_tty)
+	tcsetattr(0, TCSANOW, &orig_termios);
 }
 
 bufchain stdout_data, stderr_data;
@@ -188,6 +410,91 @@ int from_backend(void *frontend_handle, int is_stderr,
     return osize + esize;
 }
 
+int from_backend_untrusted(void *frontend_handle, const char *data, int len)
+{
+    /*
+     * No "untrusted" output should get here (the way the code is
+     * currently, it's all diverted by FLAG_STDERR).
+     */
+    assert(!"Unexpected call to from_backend_untrusted()");
+    return 0; /* not reached */
+}
+
+int get_userpass_input(prompts_t *p, unsigned char *in, int inlen)
+{
+    int ret;
+    ret = cmdline_get_passwd_input(p, in, inlen);
+    if (ret == -1)
+	ret = console_get_userpass_input(p, in, inlen);
+    return ret;
+}
+
+/*
+ * Handle data from a local tty in PARMRK format.
+ */
+static void from_tty(void *buf, unsigned len)
+{
+    char *p, *q, *end;
+    static enum {NORMAL, FF, FF00} state = NORMAL;
+
+    p = buf; end = buf + len;
+    while (p < end) {
+	switch (state) {
+	    case NORMAL:
+		if (*p == '\xff') {
+		    p++;
+		    state = FF;
+		} else {
+		    q = memchr(p, '\xff', end - p);
+		    if (q == NULL) q = end;
+		    back->send(backhandle, p, q - p);
+		    p = q;
+		}
+		break;
+	    case FF:
+		if (*p == '\xff') {
+		    back->send(backhandle, p, 1);
+		    p++;
+		    state = NORMAL;
+		} else if (*p == '\0') {
+		    p++;
+		    state = FF00;
+		} else abort();
+		break;
+	    case FF00:
+		if (*p == '\0') {
+		    back->special(backhandle, TS_BRK);
+		} else {
+		    /* 
+		     * Pretend that PARMRK wasn't set.  This involves
+		     * faking what INPCK and IGNPAR would have done if
+		     * we hadn't overridden them.  Unfortunately, we
+		     * can't do this entirely correctly because INPCK
+		     * distinguishes between framing and parity
+		     * errors, but PARMRK format represents both in
+		     * the same way.  We assume that parity errors are
+		     * more common than framing errors, and hence
+		     * treat all input errors as being subject to
+		     * INPCK.
+		     */
+		    if (orig_termios.c_iflag & INPCK) {
+			/* If IGNPAR is set, we throw away the character. */
+			if (!(orig_termios.c_iflag & IGNPAR)) {
+			    /* PE/FE get passed on as NUL. */
+			    *p = 0;
+			    back->send(backhandle, p, 1);
+			}
+		    } else {
+			/* INPCK not set.  Assume we got a parity error. */
+			back->send(backhandle, p, 1);
+		    }
+		}
+		p++;
+		state = NORMAL;
+	}
+    }
+}
+
 int signalpipe[2];
 
 void sigwinch(int signum)
@@ -236,9 +543,13 @@ static void usage(void)
     printf("  -4 -6     force use of IPv4 or IPv6\n");
     printf("  -C        enable compression\n");
     printf("  -i key    private key file for authentication\n");
+    printf("  -noagent  disable use of Pageant\n");
+    printf("  -agent    enable use of Pageant\n");
     printf("  -m file   read remote command(s) from file\n");
     printf("  -s        remote command is an SSH subsystem (SSH-2 only)\n");
     printf("  -N        don't start a shell/command (SSH-2 only)\n");
+    printf("  -nc host:port\n");
+    printf("            open tunnel in place of session (SSH-2 only)\n");
     exit(1);
 }
 
@@ -259,11 +570,9 @@ int main(int argc, char **argv)
     int exitcode;
     int errors;
     int use_subsystem = 0;
-    void *ldisc, *logctx;
+    void *ldisc;
     long now;
 
-    ssh_get_line = console_get_line;
-
     fdlist = NULL;
     fdcount = fdsize = 0;
     /*
@@ -336,11 +645,9 @@ int main(int argc, char **argv)
 		errors = 1;
 	    }
 	} else if (*p) {
-	    if (!*cfg.host) {
+	    if (!cfg_launchable(&cfg)) {
 		char *q = p;
 
-                do_defaults(NULL, &cfg);
-
 		/*
 		 * If the hostname starts with "telnet:", set the
 		 * protocol to Telnet and process the string as a
@@ -412,7 +719,7 @@ int main(int argc, char **argv)
 		    {
 			Config cfg2;
 			do_defaults(host, &cfg2);
-			if (loaded_session || cfg2.host[0] == '\0') {
+			if (loaded_session || !cfg_launchable(&cfg2)) {
 			    /* No settings for this host; use defaults */
 			    /* (or session was already loaded with -load) */
 			    strncpy(cfg.host, host, sizeof(cfg.host) - 1);
@@ -466,7 +773,7 @@ int main(int argc, char **argv)
     if (errors)
 	return 1;
 
-    if (!*cfg.host) {
+    if (!cfg_launchable(&cfg)) {
 	usage();
     }
 
@@ -522,7 +829,7 @@ int main(int argc, char **argv)
 	cfg.host[p1] = '\0';
     }
 
-    if (!cfg.remote_cmd_ptr && !*cfg.remote_cmd)
+    if (!cfg.remote_cmd_ptr && !*cfg.remote_cmd && !*cfg.ssh_nc_host)
 	flags |= FLAG_INTERACTIVE;
 
     /*
@@ -590,7 +897,7 @@ int main(int argc, char **argv)
      * fails, because we know we aren't necessarily running in a
      * console.
      */
-    tcgetattr(0, &orig_termios);
+    local_tty = (tcgetattr(0, &orig_termios) == 0);
     atexit(cleanup_termios);
     ldisc_update(NULL, 1, 1);
     sending = FALSE;
@@ -610,7 +917,7 @@ int main(int argc, char **argv)
 	FD_SET_MAX(signalpipe[0], maxfd, rset);
 
 	if (connopen && !sending &&
-	    back->socket(backhandle) != NULL &&
+	    back->connected(backhandle) &&
 	    back->sendok(backhandle) &&
 	    back->sendbuffer(backhandle) < MAX_STDIN_BACKLOG) {
 	    /* If we're OK to send, then try to read from stdin. */
@@ -725,7 +1032,7 @@ int main(int argc, char **argv)
 	    char buf[4096];
 	    int ret;
 
-	    if (connopen && back->socket(backhandle) != NULL) {
+	    if (connopen && back->connected(backhandle)) {
 		ret = read(0, buf, sizeof(buf));
 		if (ret < 0) {
 		    perror("stdin: read");
@@ -734,7 +1041,10 @@ int main(int argc, char **argv)
 		    back->special(backhandle, TS_EOF);
 		    sending = FALSE;   /* send nothing further after this */
 		} else {
-		    back->send(backhandle, buf, ret);
+		    if (local_tty)
+			from_tty(buf, ret);
+		    else
+			back->send(backhandle, buf, ret);
 		}
 	    }
 	}
@@ -747,7 +1057,7 @@ int main(int argc, char **argv)
 	    try_output(1);
 	}
 
-	if ((!connopen || back->socket(backhandle) == NULL) &&
+	if ((!connopen || !back->connected(backhandle)) &&
 	    bufchain_size(&stdout_data) == 0 &&
 	    bufchain_size(&stderr_data) == 0)
 	    break;		       /* we closed the connection */
